﻿using System;
using System.Runtime.InteropServices;

namespace Markos.IO.WindowsHooks {

	/// <summary>Class for event based keyboard hook.</summary>
	public class KeyboardHook {

		private readonly static KeyboardHook _instance = new KeyboardHook();

		/// <summary>Gets the singleton instance.</summary>
		public static KeyboardHook Instance { get { return _instance; } }

		private EventHandler<KeyStateChangedEventArgs> _keyStateChanged;

		/// <summary>Occurs when a key state changes.</summary>
		public event EventHandler<KeyStateChangedEventArgs> KeyStateChanged {
			add {
				lock(_instance) {
					if(_keyStateChanged == null)
						_windowsHook.HookInvoked += OnHookInvoked;
					_keyStateChanged += value;
				}
			}
			remove {
				lock(_instance) {
					_keyStateChanged -= value;
					if(_keyStateChanged == null) {
						_windowsHook.HookInvoked -= OnHookInvoked;
					}
				}
			}
		}

		private bool _stateAlt;
		private bool _stateCtrl;
		private bool _stateShift;
		private bool _stateWin;
		private readonly WindowsHook _windowsHook = WindowsHook.Create(HookType.WH_KEYBOARD_LL);

		private KeyboardHook() { }

		private void OnHookInvoked(object sender, WindowsHookEventArgs args) {
			if(args.Code < 0)
				return;

			var copy = _keyStateChanged;
			if(copy != null) {
				KBDLLHOOKSTRUCT data = Marshal.PtrToStructure(args.LParam, typeof(KBDLLHOOKSTRUCT)) as KBDLLHOOKSTRUCT;

				uint vkCode     = data.vkCode;
				bool down       = (WM_KeyState)args.WParam == WM_KeyState.WM_KEYDOWN || (WM_KeyState)args.WParam == WM_KeyState.WM_SYSKEYDOWN;
				bool isModifier = true;

				switch((VkCodes)vkCode) {
					case VkCodes.LAlt:
					case VkCodes.RAlt:
					case VkCodes.Alt:
						_stateAlt = down;
						break;
					case VkCodes.LShift:
					case VkCodes.RShift:
					case VkCodes.Shift:
						_stateShift = down;
						break;
					case VkCodes.LControl:
					case VkCodes.RControl:
					case VkCodes.Control:
						_stateCtrl = down;
						break;
					case VkCodes.LWin:
					case VkCodes.RWin:
						_stateWin = down;
						break;
					default:
						isModifier = false;
						break;
				}

				KeyStateChangedEventArgs eventArgs = new KeyStateChangedEventArgs(_stateAlt, _stateCtrl, _stateShift, _stateWin, down, vkCode, isModifier);

				copy(this, eventArgs);
				if(eventArgs.BlockEvent)
					args.Return = (IntPtr)1;
			}
		}

		[StructLayout(LayoutKind.Sequential)]
		private class KBDLLHOOKSTRUCT {

			public uint vkCode;
			public uint scanCode;
			public KBDLLHOOKSTRUCTFlags flags;
			public uint time;
			public UIntPtr dwExtraInfo;
		}

		[Flags()]
		private enum KBDLLHOOKSTRUCTFlags : uint {
			LLKHF_EXTENDED = 0x01,
			LLKHF_INJECTED = 0x10,
			LLKHF_ALTDOWN = 0x20,
			LLKHF_UP = 0x80,
		}

		private enum VkCodes {
			Shift = 0x10,
			LShift = 0xA0,
			RShift = 0xA1,
			Alt = 0x12,
			LAlt = 0xA4,
			RAlt = 0xA5,
			Control = 0x11,
			LControl = 0xA2,
			RControl = 0xA3,
			LWin = 0x5B,
			RWin = 0x5C,
		}

		private enum WM_KeyState : int {
			WM_KEYDOWN = 0x0100,
			WM_KEYUP = 0x0101,
			WM_SYSKEYDOWN = 0x0104,
			WM_SYSKEYUP = 0x0105,
		}
	}
}
