﻿using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;

namespace Markos.IO.WindowsHooks {

	/// <summary>Class for event based Windows hooks.</summary>
	[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1049:TypesThatOwnNativeResourcesShouldBeDisposable")]
	public class WindowsHook {

		private readonly static Dictionary<HookType, WindowsHook> _instances = new Dictionary<HookType, WindowsHook>();

		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Reliability", "CA2006:UseSafeHandleToEncapsulateNativeResources")]
		private IntPtr _hook;

		private EventHandler<WindowsHookEventArgs> _hookInvoked;

		/// <summary>Occurs when registered hook event happens.</summary>
		public event EventHandler<WindowsHookEventArgs> HookInvoked {
			add {
				lock(_hookProcCallback) {
					if(_hook == IntPtr.Zero)
						_hook = SetWindowsHookEx(_hookType, _hookProcCallback, IntPtr.Zero, 0);
					_hookInvoked += value;
				}
			}
			remove {
				lock(_hookProcCallback) {
					_hookInvoked -= value;
					if(_hookInvoked == null && _hook != IntPtr.Zero) {
						UnhookWindowsHookEx(_hook);
						_hook = IntPtr.Zero;
					}
				}
			}
		}

		/// <summary>Reference holder to avoid delegate garbage collection</summary>
		private readonly HookProc _hookProcCallback;

		private readonly HookType _hookType;

		private delegate IntPtr HookProc(int code, IntPtr wParam, IntPtr lParam);

		private WindowsHook(HookType hookType) {
			_hookProcCallback = HookProcCallback;
			_hookType = hookType;
		}

		/// <summary>Returns the singleton instance of the specified hook type.</summary>
		/// <param name="hookType">Type of the hook to register.</param>
		/// <returns>Singleton instance for the requested hook.</returns>
		public static WindowsHook Create(HookType hookType) {
			WindowsHook result;
			if(!_instances.TryGetValue(hookType, out result)) {
				result = new WindowsHook(hookType);
				_instances.Add(hookType, result);
			}
			return result;
		}

		private IntPtr HookProcCallback(int code, IntPtr wParam, IntPtr lParam) {
			var copy = _hookInvoked;
			if(copy != null) {
				var args = new WindowsHookEventArgs(code, wParam, lParam);
				_hookInvoked(this, args);
				if(args.Return.HasValue)
					return args.Return.Value;
			}
			return CallNextHookEx(_hook, code, wParam, lParam);
		}

		[DllImport("user32.dll")]
		private static extern IntPtr CallNextHookEx(IntPtr hhk, int nCode, IntPtr wParam, IntPtr lParam);

		[DllImport("user32.dll", SetLastError = true)]
		private static extern IntPtr SetWindowsHookEx(HookType hookType, HookProc lpfn, IntPtr hMod, uint dwThreadId);

		[DllImport("user32.dll", SetLastError = true)]
		[return: MarshalAs(UnmanagedType.Bool)]
		private static extern bool UnhookWindowsHookEx(IntPtr hhk);
	}
}